@@ -1,6 +1,8 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
+import warnings
+from typing import Optional
 
 try:
     from pygments import highlight
@@ -10,48 +12,53 @@ except ModuleNotFoundError:
     # Only needed if using syntax highlighting
     pass
 
-from django.core.exceptions import ImproperlyConfigured
-
 from .base import TemplateWidget, Block
 
 
-class CodeWidget(TemplateWidget):
-    """Extends TemplateWidget.
-
-    A widget to display code with context-sensitive if a language is supplied.
-
-    Args:
-        code (str): the code to be displayed
-        language (str): the language of the code
+class CodeWidget(Block):
     """
-    template_name = 'wildewidgets/code_widget.html'
-    language = None
-    code = ""
-    Line_numbers = False
-    css_class = None
+    Extends :py:class:`wildewidgets.widgets.base.Block`.
 
-    def __init__(self, *args, **kwargs):
-        if 'code' in kwargs:
-            self.code = kwargs['code']
-        if 'language' in kwargs:
-            self.language = kwargs['language']
-        self.css_class = kwargs.get("css_class", self.css_class)
+    A widget to display code with syntax highlighting if a language is supplied.
 
-    def get_context_data(self, **kwargs):
-        kwargs = super().get_context_data(**kwargs)
+    Keyword Args:
+        code: the code to be displayed
+        language: the language of the code
+    """
+    block: str = 'wildewidgets_highlight_container'
+
+    language: Optional[str] = None
+    code: str = ""
+    line_numbers: bool = False
+
+    def __init__(
+        self,
+        code: str = None,
+        language: str = None,
+        line_numbers: bool = False,
+        **kwargs
+    ):
+        self.code = code if code else self.code
+        self.language = language if language else self.language
         if not self.language:
-            raise ImproperlyConfigured("You must define a language.")
-        lexer = get_lexer_by_name(self.language)
-        formatter = HtmlFormatter(linenos=self.Line_numbers, cssclass="wildewidgets_highlight")
-        kwargs['code'] = highlight(self.code, lexer, formatter)
-        kwargs['css_class'] = self.css_class
-        return kwargs
+            raise ValueError(
+                f'{self.__class__.__name__}: "language" must be defined either as a class attribute or a keyword arg'
+            )
+        self.line_numbers = line_numbers if line_numbers else self.line_numbers
+        super().__init__(**kwargs)
+        self.add_code(self.code, language=self.language, line_numbers=self.line_numbers)
+
+    def add_code(self, code: str, language: str, line_numbers: bool = False) -> None:
+        lexer = get_lexer_by_name(language)
+        formatter = HtmlFormatter(linenos=line_numbers, cssclass="wildewidgets_highlight")
+        self.add_block(highlight(code, lexer, formatter))
 
 
 class MarkdownWidget(TemplateWidget):
-    """Extends TemplateWidget.
+    """
+    Extends :py:class:`wildewidgets.widgets.base.TemplateWidget`.
 
-    A widget to display markdown as HTML. 
+    A widget to display markdown as HTML.
 
     Args:
         text (str): the markdown to render as HTML.
@@ -62,6 +69,7 @@ class MarkdownWidget(TemplateWidget):
     css_class = ""
 
     def __init__(self, *args, **kwargs):
+        super().__init__()
         self.text = kwargs.get("text", self.text)
         self.css_class = kwargs.get("css_class", self.css_class)
 
@@ -73,9 +81,10 @@ class MarkdownWidget(TemplateWidget):
 
 
 class HTMLWidget(TemplateWidget):
-    """Extends TemplateWidget.
+    """
+    Extends :py:class:`wildewidgets.widgets.base.TemplateWidget`.
 
-    A widget to display raw HTML. 
+    A widget to display raw HTML.
 
     Args:
         html (str): the HTML to render.
@@ -97,48 +106,62 @@ class HTMLWidget(TemplateWidget):
 
 
 class StringBlock(Block):
-    """Extends Block.
+    """
+    Extends :py:class:`wildewidgets.widgets.base.Block`.
 
     A basic widget that displays a string.
 
+    .. deprecated:: 0.14.0
+
+        Use :py:class:`wildewidgets.widgets.base.Block` directly instead.  It
+        works exactly like :py:class:`StringBlock`
+
     Args:
         text: the text to display.
     """
 
     def __init__(self, text: str, **kwargs):
+        warnings.warn(
+            'Deprecated in 0.14.0; use Block directly instead.',
+            DeprecationWarning,
+            stacklevel=2
+        )
         super().__init__(*[text], **kwargs)
 
 
-class TimeStamp(StringBlock):
-    """Extends StringBlock.
+class TimeStamp(Block):
+    """
+    Extends :py:class:`wildewidgets.widgets.base.Block`.
 
     A basic widget that displays a timestamp.
 
     Args:
         text: the text to display.
     """
-    tag="small"
-    css_class="fw-light"
+    tag = "small"
+    css_class = "fw-light"
 
 
-class LabelBlock(StringBlock):
-    """Extends StringBlock.
+class LabelBlock(Block):
+    """
+    Extends :py:class:`wildewidgets.widgets.base.Block`.
 
-    A basic widget that displays a label.
+    A ``<label>``.
 
     Args:
         text: the text to display.
     """
+    tag: str = "label"
 
     def __init__(self, text: str, color: str = "secondary", **kwargs):
-        css_class = kwargs.get("css_class", "")
-        css_class += f" fw-bold"
-        kwargs["css_class"] = css_class.strip()
+        # FIXME: color is a kwarg, but it is not used
         super().__init__(text, **kwargs)
+        self.add_class('fw-bold')
 
 
-class TagBlock(StringBlock):
-    """Extends StringBlock.
+class TagBlock(Block):
+    """
+    Extends :py:class:`wildewidgets.widgets.base.Block`.
 
     A basic widget that displays a colored tag.
 
@@ -146,9 +169,8 @@ class TagBlock(StringBlock):
         text: the text to display.
         color: the bootstrap color class.
     """
-    def __init__(self, text: str, color: str = "secondary", **kwargs):
-        css_class = kwargs.get("css_class", "")
-        css_class += f" badge bg-{color}"
-        kwargs["css_class"] = css_class.strip()
-        super().__init__(text, **kwargs)
+    block: str = 'badge'
 
+    def __init__(self, text: str, color: str = "secondary", **kwargs):
+        super().__init__(text, **kwargs)
+        self.add_class(f'bg-{color}')
